require 'fileutils'
require 'test/unit'
require 'ruby_toggle_file'
require 'tmpdir'

class TestRubyToggleFile < Test::Unit::TestCase
  WORK_DIR = "#{Dir.tmpdir}/zdsfwfwejiotest".freeze
  FileUtils.rm_rf WORK_DIR

  FILE_MAP = {
    :zero  => %w[lib/zero.rb     test/test_zero.rb],
    :one   => %w[lib/one/one.rb  test/one/test_one.rb],
    :two   => %w[lib/two/two.rb  test/test_two.rb],
    :three => %w[three.rb        test_three.rb],
    :four  => %w[four.rb],
    :five  => %w[lib/five.rb],
    :six   => %w[lib/six/six/six.rb],
    :seven => %w[                test_seven.rb],
    :eight => %w[                test/test_eight.rb],
    :nine  => %w[                test/nine/nine/nine.rb],

    # Rails
    :rcontrollers => %w[app/controllers/c.rb   test/functional/c_test.rb],
    :rmodels      => %w[app/models/m.rb        test/unit/m_test.rb],
    :rlib         => %w[lib/l.rb               test/unit/test_l.rb app/models/m.rb],
  }

  def use_file_map(key, &block)
    Dir.mkdir(WORK_DIR)
    Dir.chdir(WORK_DIR) do
      FILE_MAP[key].each do |file|
        FileUtils.mkpath(File.dirname(file))
        open(file, "w"){}
      end
    end

    begin
      yield
    ensure
      FileUtils.rm_rf WORK_DIR
    end
  end

  def setup
    @x = RubyToggleFile.new
  end

  def _(path)
    WORK_DIR + "/" + path
  end

  def check(after, before)
    assert_equal _(after), @x.ruby_toggle_file(_(before))
  end

  def test_test_file_file_exist
    use_file_map(:zero)  { check "test/test_zero.rb",    "lib/zero.rb"    }
    use_file_map(:one)   { check "test/one/test_one.rb", "lib/one/one.rb" }
    use_file_map(:two)   { check "test/test_two.rb",     "lib/two/two.rb" }
    use_file_map(:three) { check "test_three.rb",        "three.rb"       }
    end

  def test_test_file_file_not_exist
    use_file_map(:four) { check "test_four.rb",             "four.rb"            }
    use_file_map(:five) { check "test/test_five.rb",        "lib/five.rb"        }
    use_file_map(:six)  { check "test/six/six/test_six.rb", "lib/six/six/six.rb" }
  end

  def test_implementation_file_file_exist
    use_file_map(:zero)  { check "lib/zero.rb",    "test/test_zero.rb"    }
    use_file_map(:one)   { check "lib/one/one.rb", "test/one/test_one.rb" }
    use_file_map(:two)   { check "lib/two/two.rb", "test/test_two.rb"     }
    use_file_map(:three) { check "three.rb",       "test_three.rb"        }
  end

  def test_implementation_file_file_not_exist
    use_file_map(:seven) { check "seven.rb",     "test_seven.rb"      }
    use_file_map(:eight) { check "lib/eight.rb", "test/test_eight.rb" }
    use_file_map(:nine)  { check "lib/nine/nine/nine.rb", "test/nine/nine/test_nine.rb" }
  end

#  Rails test
  def test_test_file_rails_controllers
    use_file_map(:rcontrollers) { check "test/functional/c_test.rb", "app/controllers/c.rb" }
  end

  def test_test_file_rails_models
    use_file_map(:rmodels) { check "test/unit/m_test.rb", "app/models/m.rb" }
  end

  def test_test_file_rails_lib
    use_file_map(:rlib) { check "test/unit/test_l.rb", "lib/l.rb" }
  end


  def test_implementation_file_rails_controllers
    use_file_map(:rcontrollers) { check "app/controllers/c.rb", "test/functional/c_test.rb" }
  end

  def test_implementation_file_rails_models
    use_file_map(:rmodels) { check "app/models/m.rb", "test/unit/m_test.rb" }
  end

  def test_implementation_file_rails_lib
    use_file_map(:rlib) { check "lib/l.rb", "test/unit/test_l.rb" }
  end


end


class TestRunHooksWithArgsUntilSuccess < Test::Unit::TestCase
  def m001(x) nil end
  private
  def m002(x) false end
  def m003(x) 100*x end
  def m004(x) 200 end

  public
  def test_run_hooks_with_args_until_success
    assert_equal 1000, run_hooks_with_args_until_success(/^m\d+$/, 10)
    assert_nil   run_hooks_with_args_until_success(/^m001$/, 10)
    assert_equal 200, run_hooks_with_args_until_success(/^m004$/, 10)
  end
end
